@@ -25,7 +25,7 @@
  */
 
 /*
- *  Copyright (C) 2003-2015 Max-Planck-Society
+ *  Copyright (C) 2003-2017 Max-Planck-Society
  *  Author: Martin Reinecke
  */
 
@@ -33,7 +33,7 @@
 #include "alm.h"
 #include "healpix_map.h"
 #include "xcomplex.h"
-#include "sharp_cxx.h"
+#include "libsharp/sharp_cxx.h"
 
 using namespace std;
 
@@ -70,6 +70,25 @@ template void map2alm (const Healpix_Map<double> &map,
   Alm<xcomplex<double> > &alm, const arr<double> &weight,
   bool add_alm);
 
+template<typename T> void alm2map_adjoint (const Healpix_Map<T> &map,
+  Alm<xcomplex<T> > &alm, bool add_alm)
+  {
+  planck_assert (map.Scheme()==RING,
+    "alm2map_adjoint: map must be in RING scheme");
+  planck_assert (map.fullyDefined(),"map contains undefined pixels");
+  checkLmaxNside(alm.Lmax(), map.Nside());
+
+  sharp_cxxjob<T> job;
+  job.set_Healpix_geometry (map.Nside());
+  job.set_triangular_alm_info (alm.Lmax(), alm.Mmax());
+  job.alm2map_adjoint(&map[0], &alm(0,0), add_alm);
+  }
+
+template void alm2map_adjoint (const Healpix_Map<float> &map,
+  Alm<xcomplex<float> > &alm, bool add_alm);
+template void alm2map_adjoint (const Healpix_Map<double> &map,
+  Alm<xcomplex<double> > &alm, bool add_alm);
+
 template<typename T> void map2alm_iter (const Healpix_Map<T> &map,
   Alm<xcomplex<T> > &alm, int num_iter, const arr<double> &weight)
   {
@@ -125,6 +144,7 @@ template<typename T> void map2alm_spin
    Alm<xcomplex<T> > &alm1, Alm<xcomplex<T> > &alm2,
    int spin, const arr<double> &weight, bool add_alm)
   {
+  planck_assert (spin>0, "map2alm_spin: spin must be positive");
   planck_assert (map1.Scheme()==RING,
     "map2alm_spin: maps must be in RING scheme");
   planck_assert (map1.conformable(map2),
@@ -152,11 +172,44 @@ template void map2alm_spin
    Alm<xcomplex<double> > &alm1, Alm<xcomplex<double> > &alm2,
    int spin, const arr<double> &weight, bool add_alm);
 
+template<typename T> void alm2map_spin_adjoint
+  (const Healpix_Map<T> &map1, const Healpix_Map<T> &map2,
+   Alm<xcomplex<T> > &alm1, Alm<xcomplex<T> > &alm2,
+   int spin, bool add_alm)
+  {
+  planck_assert (spin>0, "alm2map_spin_adjoint: spin must be positive");
+  planck_assert (map1.Scheme()==RING,
+    "alm2map_spin_adjoint: maps must be in RING scheme");
+  planck_assert (map1.conformable(map2),
+    "alm2map_spin_adjoint: maps are not conformable");
+  planck_assert (alm1.conformable(alm1),
+    "alm2map_spin_adjoint: a_lm are not conformable");
+  planck_assert (map1.fullyDefined()&&map2.fullyDefined(),
+    "map contains undefined pixels");
+  checkLmaxNside(alm1.Lmax(), map1.Nside());
+
+  sharp_cxxjob<T> job;
+  job.set_Healpix_geometry (map1.Nside());
+  job.set_triangular_alm_info (alm1.Lmax(), alm1.Mmax());
+  job.alm2map_spin_adjoint(&map1[0],&map2[0],&alm1(0,0),&alm2(0,0),spin,
+    add_alm);
+  }
+
+template void alm2map_spin_adjoint
+  (const Healpix_Map<float> &map1, const Healpix_Map<float> &map2,
+   Alm<xcomplex<float> > &alm1, Alm<xcomplex<float> > &alm2,
+   int spin, bool add_alm);
+template void alm2map_spin_adjoint
+  (const Healpix_Map<double> &map1, const Healpix_Map<double> &map2,
+   Alm<xcomplex<double> > &alm1, Alm<xcomplex<double> > &alm2,
+   int spin, bool add_alm);
+
 template<typename T> void map2alm_spin_iter2
   (const Healpix_Map<T> &map1, const Healpix_Map<T> &map2,
    Alm<xcomplex<T> > &alm1, Alm<xcomplex<T> > &alm2,
    int spin, double err_abs, double err_rel)
   {
+  planck_assert (spin>0, "map2alm_spin: spin must be positive");
   arr<double> wgt(2*map1.Nside());
   wgt.fill(1);
   Healpix_Map<T> map1b(map1), map2b(map2);
@@ -235,6 +288,48 @@ template void map2alm_pol
    const arr<double> &weight,
    bool add_alm);
 
+template<typename T> void alm2map_pol_adjoint
+  (const Healpix_Map<T> &mapT,
+   const Healpix_Map<T> &mapQ,
+   const Healpix_Map<T> &mapU,
+   Alm<xcomplex<T> > &almT,
+   Alm<xcomplex<T> > &almG,
+   Alm<xcomplex<T> > &almC,
+   bool add_alm)
+  {
+  planck_assert (mapT.Scheme()==RING,
+    "alm2map_pol_adjoint: maps must be in RING scheme");
+  planck_assert (mapT.conformable(mapQ) && mapT.conformable(mapU),
+    "alm2map_pol_adjoint: maps are not conformable");
+  planck_assert (almT.conformable(almG) && almT.conformable(almC),
+    "alm2map_pol_adjoint: a_lm are not conformable");
+  planck_assert (mapT.fullyDefined()&&mapQ.fullyDefined()&&mapU.fullyDefined(),
+    "map contains undefined pixels");
+  checkLmaxNside(almT.Lmax(), mapT.Nside());
+
+  sharp_cxxjob<T> job;
+  job.set_Healpix_geometry (mapT.Nside());
+  job.set_triangular_alm_info (almT.Lmax(), almT.Mmax());
+  job.alm2map_adjoint(&mapT[0], &almT(0,0), add_alm);
+  job.alm2map_spin_adjoint(&mapQ[0],&mapU[0],&almG(0,0),&almC(0,0),2,add_alm);
+  }
+
+template void alm2map_pol_adjoint
+  (const Healpix_Map<float> &mapT,
+   const Healpix_Map<float> &mapQ,
+   const Healpix_Map<float> &mapU,
+   Alm<xcomplex<float> > &almT,
+   Alm<xcomplex<float> > &almG,
+   Alm<xcomplex<float> > &almC,
+   bool add_alm);
+template void alm2map_pol_adjoint
+  (const Healpix_Map<double> &mapT,
+   const Healpix_Map<double> &mapQ,
+   const Healpix_Map<double> &mapU,
+   Alm<xcomplex<double> > &almT,
+   Alm<xcomplex<double> > &almG,
+   Alm<xcomplex<double> > &almC,
+   bool add_alm);
 
 template<typename T> void map2alm_pol_iter
   (const Healpix_Map<T> &mapT,
@@ -332,25 +427,26 @@ template void map2alm_pol_iter2
 
 
 template<typename T> void alm2map (const Alm<xcomplex<T> > &alm,
-  Healpix_Map<T> &map)
+  Healpix_Map<T> &map, bool add_map)
   {
   planck_assert (map.Scheme()==RING, "alm2map: map must be in RING scheme");
 
   sharp_cxxjob<T> job;
   job.set_Healpix_geometry (map.Nside());
   job.set_triangular_alm_info (alm.Lmax(), alm.Mmax());
-  job.alm2map(&alm(0,0), &map[0], false);
+  job.alm2map(&alm(0,0), &map[0], add_map);
   }
 
 template void alm2map (const Alm<xcomplex<double> > &alm,
-  Healpix_Map<double> &map);
+  Healpix_Map<double> &map, bool add_map);
 template void alm2map (const Alm<xcomplex<float> > &alm,
-  Healpix_Map<float> &map);
+  Healpix_Map<float> &map, bool add_map);
 
 template<typename T> void alm2map_spin
   (const Alm<xcomplex<T> > &alm1, const Alm<xcomplex<T> > &alm2,
-   Healpix_Map<T> &map1, Healpix_Map<T> &map2, int spin)
+   Healpix_Map<T> &map1, Healpix_Map<T> &map2, int spin, bool add_map)
   {
+  planck_assert (spin>0, "alm2map_spin: spin must be positive");
   planck_assert (map1.Scheme()==RING,
     "alm2map_spin: maps must be in RING scheme");
   planck_assert (map1.conformable(map2),
@@ -361,15 +457,15 @@ template<typename T> void alm2map_spin
   sharp_cxxjob<T> job;
   job.set_Healpix_geometry (map1.Nside());
   job.set_triangular_alm_info (alm1.Lmax(), alm1.Mmax());
-  job.alm2map_spin(&alm1(0,0),&alm2(0,0),&map1[0],&map2[0],spin,false);
+  job.alm2map_spin(&alm1(0,0),&alm2(0,0),&map1[0],&map2[0],spin,add_map);
   }
 
 template void alm2map_spin
   (const Alm<xcomplex<double> > &alm1, const Alm<xcomplex<double> > &alm2,
-   Healpix_Map<double> &map, Healpix_Map<double> &map2, int spin);
+   Healpix_Map<double> &map, Healpix_Map<double> &map2, int spin, bool add_map);
 template void alm2map_spin
   (const Alm<xcomplex<float> > &alm1, const Alm<xcomplex<float> > &alm2,
-   Healpix_Map<float> &map, Healpix_Map<float> &map2, int spin);
+   Healpix_Map<float> &map, Healpix_Map<float> &map2, int spin, bool add_map);
 
 
 template<typename T> void alm2map_pol
@@ -378,7 +474,8 @@ template<typename T> void alm2map_pol
    const Alm<xcomplex<T> > &almC,
    Healpix_Map<T> &mapT,
    Healpix_Map<T> &mapQ,
-   Healpix_Map<T> &mapU)
+   Healpix_Map<T> &mapU,
+   bool add_map)
   {
   planck_assert (mapT.Scheme()==RING,
     "alm2map_pol: maps must be in RING scheme");
@@ -390,8 +487,8 @@ template<typename T> void alm2map_pol
   sharp_cxxjob<T> job;
   job.set_Healpix_geometry (mapT.Nside());
   job.set_triangular_alm_info (almT.Lmax(), almT.Mmax());
-  job.alm2map(&almT(0,0), &mapT[0], false);
-  job.alm2map_spin(&almG(0,0), &almC(0,0), &mapQ[0], &mapU[0], 2, false);
+  job.alm2map(&almT(0,0), &mapT[0], add_map);
+  job.alm2map_spin(&almG(0,0), &almC(0,0), &mapQ[0], &mapU[0], 2, add_map);
   }
 
 template void alm2map_pol (const Alm<xcomplex<double> > &almT,
@@ -399,14 +496,16 @@ template void alm2map_pol (const Alm<xcomplex<double> > &almT,
                            const Alm<xcomplex<double> > &almC,
                            Healpix_Map<double> &mapT,
                            Healpix_Map<double> &mapQ,
-                           Healpix_Map<double> &mapU);
+                           Healpix_Map<double> &mapU,
+                           bool add_map);
 
 template void alm2map_pol (const Alm<xcomplex<float> > &almT,
                            const Alm<xcomplex<float> > &almG,
                            const Alm<xcomplex<float> > &almC,
                            Healpix_Map<float> &mapT,
                            Healpix_Map<float> &mapQ,
-                           Healpix_Map<float> &mapU);
+                           Healpix_Map<float> &mapU,
+                           bool add_map);
 
 
 template<typename T> void alm2map_der1
