@@ -12,5 +12,7 @@ int mult_alm_module (int argc, const char **argv);
 int smoothing_cxx_module (int argc, const char **argv);
 int udgrade_cxx_module (int argc, const char **argv);
 int udgrade_harmonic_cxx_module (int argc, const char **argv);
+int compute_weights_module (int argc, const char **argv);
+int needlet_tool (int argc, const char **argv);
 
 #endif
